package fe

import (
	"fmt"
	"html/template"
	"net/http"
	"sort"
	"sync"
	"time"
)

const debugText = `<html>
    <body>
    <title>Requests</title>
    <table>
      <tr><th>IP</th><th>URL</th><th>Target</th></tr>
      {{range .}}
      <tr>
        <td align=left>{{.RemoteAddr}}</td>
        <td align=left font=fixed>{{.URL}}</td>
        <td align=left>{{.State}}</td>
        <td align=left>{{.Target}}</td>
      </tr>  
      {{end}}
    </table>
    </body>
    </html>`

var (
	debugTmpl = template.Must(template.New("http debug").Parse(debugText))

	tracker *requestTracker
)

const (
	StateConnected = iota
	StateSource
	StateListener
)

type connState int

func (s connState) String() string {
	switch s {
	case StateConnected:
		return "CONNECTED"
	case StateSource:
		return "SOURCE"
	case StateListener:
		return "LISTENER"
	}
	return ""
}

type connInfo struct {
	State      connState
	URL        string
	Stamp      time.Time
	RemoteAddr string
	Target     string
}

// Request tracker (for long-running requests).
type requestTracker struct {
	lock  sync.Mutex
	conns map[*http.Request]*connInfo
}

func (t *requestTracker) addConn(r *http.Request) {
	t.lock.Lock()
	defer t.lock.Unlock()
	t.conns[r] = &connInfo{
		URL:        r.URL.String(),
		State:      StateConnected,
		Stamp:      time.Now(),
		RemoteAddr: r.RemoteAddr,
	}
}

func (t *requestTracker) delConn(r *http.Request) {
	t.lock.Lock()
	defer t.lock.Unlock()
	delete(t.conns, r)
}

func (t *requestTracker) setConnState(r *http.Request, state connState, target string) {
	t.lock.Lock()
	defer t.lock.Unlock()
	if cs, ok := t.conns[r]; ok {
		cs.State = state
		cs.Target = target
	}
}

type connList []connInfo

func (l connList) Len() int      { return len(l) }
func (l connList) Swap(i, j int) { l[i], l[j] = l[j], l[i] }
func (l connList) Less(i, j int) bool {
	if l[i].URL == l[j].URL {
		return l[i].RemoteAddr < l[j].RemoteAddr
	}
	return l[i].URL < l[j].URL
}

func (t *requestTracker) ServeHTTP(w http.ResponseWriter, r *http.Request) {
	t.lock.Lock()
	var conns []connInfo
	for _, c := range t.conns {
		conns = append(conns, *c)
	}
	t.lock.Unlock()

	sort.Sort(connList(conns))

	err := debugTmpl.Execute(w, conns)
	if err != nil {
		fmt.Fprintln(w, "debug: error executing template: ", err.Error())
	}
}

func trackRequestsHandler(h http.Handler) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		tracker.addConn(r)
		h.ServeHTTP(w, r)
		tracker.delConn(r)
	})
}

func init() {
	tracker = &requestTracker{
		conns: make(map[*http.Request]*connInfo),
	}
	http.Handle("/debug/http", tracker)
}
