package fe

import (
	"fmt"
	"io/ioutil"
	"net"
	"net/http"
	"net/http/httptest"
	"strconv"
	"strings"
	"testing"

	"git.autistici.org/ale/autoradio"
	"git.autistici.org/ale/autoradio/util"
)

// Create a test target, replacing the local icecast daemon. Returns
// the port used, so the caller can then override the IcecastPort
// global variable.
func createTestTargetServer(t *testing.T) (*httptest.Server, int) {
	srv := httptest.NewServer(http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		fmt.Fprintf(w, "%s", r.URL.Path)
	}))
	port, _ := strconv.Atoi(strings.Split(strings.TrimPrefix(srv.URL, "http://"), ":")[1])
	return srv, port
}

func createTestHttpRedirector(t *testing.T) *HttpRedirector {
	nodes := []*autoradio.NodeStatus{
		&autoradio.NodeStatus{
			Name:      "node1",
			IcecastUp: true,
			IP: []net.IP{
				net.ParseIP("127.0.0.1"),
			},
		},
	}

	etcd := util.NewTestEtcdClient()
	etcd.Set(autoradio.MountPrefix+"test.ogg",
		`{"Name": "/test.ogg", "Username": "source1", "Password": "foo"}`,
		86400)
	etcd.Set(autoradio.MasterElectionPath,
		`{"Name": "node1", "IP": ["127.0.0.1"]}`,
		86400)
	client := autoradio.NewClient(etcd)
	h, err := NewHttpRedirector(client, "example.com", "best", "./static", "./templates")
	if err != nil {
		t.Fatal(err)
	}

	// Bootstrap the load balancer.
	h.lb.Update(nodes)

	return h
}

func createTestHttpServer(t *testing.T) (*HttpRedirector, *httptest.Server) {
	h := createTestHttpRedirector(t)
	srv := httptest.NewServer(h.createHandler())
	return h, srv
}

func doHttpRequest(t *testing.T, method, url string, expectedStatus int) string {
	req, err := http.NewRequest(method, url, nil)
	if err != nil {
		t.Fatalf("Error building request: %v", err)
	}
	client := &http.Client{}
	resp, err := client.Do(req)
	if err != nil {
		t.Fatalf("Error retrieving page: %v", err)
	}
	if resp.StatusCode != expectedStatus {
		t.Fatalf("HTTP status %d (expected %d)", resp.StatusCode, expectedStatus)
	}
	defer resp.Body.Close()
	data, err := ioutil.ReadAll(resp.Body)
	if err != nil {
		t.Fatalf("Error reading data: %v", err)
	}
	return string(data)
}

func TestHttpRedirector_StatusPage(t *testing.T) {
	_, srv := createTestHttpServer(t)
	defer srv.Close()

	// Retrieve the status page.
	data := doHttpRequest(t, "GET", srv.URL, 200)
	if !strings.Contains(data, "<div class=\"container\">") {
		t.Errorf("Bad response:\n%s", data)
	}
}

func TestHttpRedirector_Static(t *testing.T) {
	_, srv := createTestHttpServer(t)
	defer srv.Close()

	data := doHttpRequest(t, "GET", srv.URL+"/static/style.css", 200)
	if !strings.Contains(data, "body {") {
		t.Errorf("Bad response:\n%s", data)
	}
}

func TestHttpRedirector_LBDebugPage(t *testing.T) {
	_, srv := createTestHttpServer(t)
	defer srv.Close()

	data := doHttpRequest(t, "GET", srv.URL+"/debug/lbv2", 200)
	if !strings.Contains(data, "node1") {
		t.Errorf("Bad response:\n%s", data)
	}
}

type httpTestContext struct {
	srv             *httptest.Server
	targetSrv       *httptest.Server
	origIcecastPort int
}

func (c *httpTestContext) Close() {
	c.srv.Close()
	c.targetSrv.Close()
	autoradio.IcecastPort = c.origIcecastPort
}

func createTestHttpContext(t *testing.T) *httpTestContext {
	target, targetPort := createTestTargetServer(t)
	autoradio.IcecastPort = targetPort
	_, srv := createTestHttpServer(t)
	c := &httpTestContext{
		srv:             srv,
		targetSrv:       target,
		origIcecastPort: autoradio.IcecastPort,
	}
	return c
}

func TestHttpRedirector_Source(t *testing.T) {
	ctx := createTestHttpContext(t)
	defer ctx.Close()

	data := doHttpRequest(t, "SOURCE", ctx.srv.URL+"/test.ogg", 200)
	if data != "/_stream/test.ogg" {
		t.Errorf("Bad response: %s", data)
	}

	doHttpRequest(t, "SOURCE", ctx.srv.URL+"/nonexist.ogg", 404)
}

func TestHttpRedirector_Relay(t *testing.T) {
	ctx := createTestHttpContext(t)
	defer ctx.Close()

	data := doHttpRequest(t, "GET", ctx.srv.URL+"/test.ogg", 200)
	if data != "/_stream/test.ogg" {
		t.Errorf("Bad response: %s", data)
	}
}

func TestHttpRedirector_IcecastProxy(t *testing.T) {
	*proxyStreams = true
	defer func() {
		*proxyStreams = false
	}()

	ctx := createTestHttpContext(t)
	defer ctx.Close()

	data := doHttpRequest(t, "GET", ctx.srv.URL+"/_stream/test.ogg", 200)
	if data != "/_stream/test.ogg" {
		t.Errorf("Bad response: %s", data)
	}
}
